from .writers.batch_writer import BatchWriter
from .writers.stream_writer import StreamWriter
from .writers.writer import Writer

from .readers.reader import Reader
from .readers.internals.sql_reader import SqlReader

from .internals.dictset import STORAGE_CLASS, DictSet



"""
https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry
"""
import os
import atexit
import logging
import datetime
import orjson as json
from typing import Union, Optional, Dict
from ..logging.levels import LEVELS, LEVELS_TO_STRING
from ..utils import is_running_from_ipython
from ..logging.log_formatter import LogFormatter


logging_seen_warnings: Dict[int, int] = {}


def fix_dict(obj: dict) -> dict:
    def fix_fields(dt):
        if isinstance(dt, (datetime.date, datetime.datetime)):
            return dt.isoformat()
        if isinstance(dt, bytes):
            return dt.decode("UTF8")
        if isinstance(dt, dict):
            return {k: fix_fields(v) for k, v in dt.items()}
        return str(dt)

    if not isinstance(obj, dict):
        return obj  # type:ignore
    return {k: fix_fields(v) for k, v in obj.items()}


def report_suppressions(message):
    from .. import logging as ml

    record = logging_seen_warnings.get(hash(message))
    if record:
        ml.get_logger().warning(
            f'The following message was suppressed {record} time(s) - "{message}"'
        )


def log_it(payload):
    if isinstance(payload, dict):
        payload = json.dumps(fix_dict(payload))
    if isinstance(payload, bytes):
        payload = payload.decode()
    print(payload, flush=True)
    return payload


def extract_caller():
    import traceback
    import os.path

    frames = traceback.extract_stack()
    if len(frames) < 4:
        return "<unknown>", "<unknown>", -1
    frame = frames[len(frames) - 4]
    head, tail = os.path.split(frame.filename)
    return frame.name, tail, frame.lineno


class GoogleLogger(object):
    @staticmethod
    def supported():
        if is_running_from_ipython():
            return False
        if os.environ.get("K_SERVICE", "") or "" != "":  # nosemgrep
            return True
        return False

    @staticmethod
    def write_event(
        message: Union[str, dict],
        system: Optional[str] = None,
        severity: Optional[int] = logging.DEBUG,
        spanId: Optional[str] = None,
    ):

        # supress duplicate warnings
        if severity == LEVELS.WARNING:  # warnings
            hashed = hash(str(message))
            if hashed in logging_seen_warnings:
                logging_seen_warnings[hashed] += 1
                return "suppressed"
            logging_seen_warnings[hashed] = 0
            atexit.register(report_suppressions, str(message))

        structured_log = {
            "severity": str(severity).split(".")[1],
            "logging.googleapis.com/labels": {
                "system": system,
                "platform": os.environ.get("LOG_SINK"),
                "formatter": "MABEL",
            },
        }

        method, module, line = extract_caller()
        structured_log["logging.googleapis.com/sourceLocation"] = {
            "function": method,
            "file": module,
            "line": line,
        }  # type:ignore

        if spanId:
            structured_log["logging.googleapis.com/spanId"] = spanId

        if isinstance(message, dict):
            formatter = LogFormatter(None)
            message = formatter.clean_record(message, False)
            structured_log["message"] = (
                str(message) + " *"
            )  # We don't want valid JSON here
            structured_log.update(message)  # type:ignore
            return log_it(structured_log)
        else:
            structured_log["message"] = message
            return log_it(structured_log)

    def __init__(self):
        # rewrite one of the levels
        LEVELS_TO_STRING[LEVELS.AUDIT] = "NOTICE"

        self.level = int(os.environ.get("LOGGING_LEVEL", 25))
        self.debug = self.create_logger(LEVELS.DEBUG)
        self.info = self.create_logger(LEVELS.INFO)
        self.warning = self.create_logger(LEVELS.WARNING)
        self.error = self.create_logger(LEVELS.ERROR)
        self.audit = self.create_logger(LEVELS.AUDIT)
        self.alert = self.create_logger(LEVELS.ALERT)

    def create_logger(self, level):
        from .create_logger import LOG_NAME

        def base_logger(message):
            return GoogleLogger.write_event(
                message=message, system=LOG_NAME, severity=level
            )

        def do_nothing(message):
            pass

        if level > self.level:
            return base_logger
        else:
            return do_nothing

    def setLevel(self, level):
        self.level = level
        self.debug = self.create_logger(LEVELS.DEBUG)
        self.info = self.create_logger(LEVELS.INFO)
        self.warning = self.create_logger(LEVELS.WARNING)
        self.error = self.create_logger(LEVELS.ERROR)
        self.audit = self.create_logger(LEVELS.AUDIT)
        self.alert = self.create_logger(LEVELS.ALERT)

    def __call__(self, message):
        self.debug(message)



pass_thru = lambda x: x


def json(ds):
    import simdjson

    """parse each line in the file to a dictionary"""
    try:
        json_parser = simdjson.Parser()
        return json_parser.parse(ds)
    except ValueError:
        # fall back
        import json

        return json.loads(ds)


def pass_thru_block(ds):
    """each blob is read as a block"""
    if isinstance(ds, str):
        return ds
    return "\n".join([r for r in ds])  # pragma: no cover


def xml(ds):
    from ...internals import xmler

    return xmler.parse(ds)



"""
This is a filtering mechanism to be applied when reading data.
"""
import operator
from typing import Optional, Iterable, List, Tuple, Union
from ...errors import InvalidSyntaxError
from ...logging import get_logger
from ...utils.text import like, matches


def _in(x, y):
    return x in y


def _nin(x, y):
    return x not in y


def _con(x, y):
    return y in x


def _ncon(x, y):
    return y not in x


def true(x):
    return True


# convert text representation of operators to functions
OPERATORS = {
    "=": operator.eq,
    "==": operator.eq,
    "is": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge,
    "like": like,
    "matches": matches,
    "in": _in,
    "!in": _nin,
    "not in": _nin,
    "contains": _con,
    "!contains": _ncon,
}


def evaluate(predicate: Union[tuple, list], record: dict) -> bool:
    """
    This is the evaluation routine for the Filter class.

    Implements a DNF (Disjunctive Normal Form) interpretter. Predicates in the same
    list are joined with an AND (*all()*) and predicates in adjacent lists are joined
    with an OR (*any()*). This allows for non-trivial filters to be written with just
    _tuples_ and _lists_.

    The predicates are in _tuples_ in the form (`key`, `op`, `value`) where the `key`
    is the value looked up from the record, the `op` is the operator and the `value`
    is a literal.
    """
    # No filter doesn't filter
    if predicate is None:  # pragma: no cover
        return True

    # If we have a tuple extract out the key, operator and value and do the evaluation
    if isinstance(predicate, tuple):
        key, op, value = predicate
        if key in record:
            return OPERATORS[op.lower()](record[key], value)
        return False

    if isinstance(predicate, list):
        # Are all of the entries tuples?
        # We AND them together (_all_ are True)
        if all([isinstance(p, tuple) for p in predicate]):
            return all([evaluate(p, record) for p in predicate])

        # Are all of the entries lists?
        # We OR them together (_any_ are True)
        if all([isinstance(p, list) for p in predicate]):
            return any([evaluate(p, record) for p in predicate])

        # if we're here the structure of the filter is wrong
        raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover

    raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover


def get_indexable_filter_columns(predicate):
    """
    Returns all of the columns in a filter which the operation benefits
    from an index

    This creates an list of tuples of (field,value) that we can feed to the
    index search.
    """
    INDEXABLE_OPS = {"=", "==", "is", "in", "contains"}
    if predicate is None:
        return []
    if isinstance(predicate, tuple):
        key, op, value = predicate
        if op in INDEXABLE_OPS:
            return [
                (
                    key,
                    value,
                )
            ]
    if isinstance(predicate, list):
        if all([isinstance(p, tuple) for p in predicate]):
            return [
                (
                    k,
                    v,
                )
                for k, o, v in predicate
                if o in INDEXABLE_OPS
            ]
        if all([isinstance(p, list) for p in predicate]):
            columns = []
            for p in predicate:
                columns += get_indexable_filter_columns(p)
            return columns
    return []  # pragma: no cover


class DnfFilters:

    __slots__ = ("empty_filter", "predicates")

    def __init__(self, filters: Optional[List[Tuple[str, str, object]]] = None):
        """
        A class to supporting filtering data.

        Parameters:
            filters: list of tuples
                Each tuple has format: (`key`, `op`, `value`). When run the
                filter will extract the `key` field from the dictionary and
                compare to the `value` using the operator `op`. Multiple
                filters are treated as AND, lists of ANDs are treated as ORs.
                The supported `op` values are: `=` or `==`, `!=`, `<`, `>`,
                `<=`, `>=`, `in`, `!in` (not in), `contains`, `!contains`
                (does not contain) and `like`. If the `op` is `in` or `!in`,
                the `value` must be a collection such as a _list_, a _set_
                or a _tuple_.
                `like` performs similar to the SQL operator, `%` is a
                multi-character wildcard and `_` is a single character
                wildcard.

        Examples:
            filters = Filters([('name', '==', 'john')])
            filters = Filters([('name', '!=', 'john'),('name', '!=', 'tom')])

        """
        if filters:
            self.predicates = filters
            self.empty_filter = False
        else:
            self.empty_filter = True

    def _get_filter_columns(self, predicate):
        if predicate is None:
            return []
        if isinstance(predicate, tuple):
            key, op, value = predicate
            return [key]
        if isinstance(predicate, list):
            if all([isinstance(p, tuple) for p in predicate]):
                return [k for k, o, v in predicate]
            if all([isinstance(p, list) for p in predicate]):
                columns = []
                for p in predicate:
                    columns += self._get_filter_columns(p)
                return columns
        return []  # pragma: no cover

    def filter_dictset(self, dictset: Iterable[dict]) -> Iterable:
        """
        Tests each entry in a Iterable against the filters

        Parameters:
            dictset: iterable of dictionaries
                The dictset to process

        Yields:
            dictionary
        """
        if self.empty_filter:
            yield from dictset
        else:
            for record in dictset:
                if evaluate(self.predicates, record):
                    yield record

    def __call__(self, record) -> bool:
        return evaluate(self.predicates, record)



import threading
import orjson
import json
import zstandard
from ...internals.index import IndexBuilder
from ...internals.records import flatten
from ....logging import get_logger
from ....utils.paths import get_parts
from ....utils import safe_field_name
from ....errors import MissingDependencyError


BLOB_SIZE = 64 * 1024 * 1024  # 64Mb, 16 files per gigabyte
SUPPORTED_FORMATS_ALGORITHMS = ("jsonl", "zstd", "parquet", "text", "flat")


class BlobWriter(object):

    # in som failure scenarios commit is called before __init__, so we need to define
    # this variable outside the __init__.
    buffer = bytearray()

    def __init__(
        self,
        *,  # force params to be named
        inner_writer=None,  # type:ignore
        blob_size: int = BLOB_SIZE,
        format: str = "zstd",
        **kwargs,
    ):

        self.format = format
        self.maximum_blob_size = blob_size

        if format not in SUPPORTED_FORMATS_ALGORITHMS:
            raise ValueError(
                f"Invalid format `{format}`, valid options are {SUPPORTED_FORMATS_ALGORITHMS}"
            )

        kwargs["format"] = format
        self.inner_writer = inner_writer(**kwargs)  # type:ignore

        self.open_buffer()

    def append(self, record: dict = {}):
        # serialize the record
        if self.format == "text":
            if isinstance(record, bytes):
                serialized = record + b"\n"
            elif isinstance(record, str):
                serialized = record.encode() + b"\n"
            else:
                serialized = str(record).encode() + b"\n"
        elif self.format == "flat":
            serialized = orjson.dumps(flatten(record)) + b"\n"  # type:ignore
        elif hasattr(record, "mini"):
            serialized = record.mini + b"\n"  # type:ignore
        else:
            try:
                serialized = orjson.dumps(record) + b"\n"  # type:ignore
            except TypeError:
                serialized = json.dumps(record).encode() + b"\n"

        # the newline isn't counted so add 1 to get the actual length
        # if this write would exceed the blob size, close it so another
        # blob will be created
        if len(self.buffer) > self.maximum_blob_size and self.records_in_buffer > 0:
            self.commit()
            self.open_buffer()

        # write the record to the file
        self.buffer.extend(serialized)
        self.records_in_buffer += 1

        return self.records_in_buffer

    def commit(self):

        committed_blob_name = ""

        if len(self.buffer) > 0:

            lock = threading.Lock()

            try:
                lock.acquire(blocking=True, timeout=10)

                if self.format == "parquet":
                    try:
                        import pyarrow.json
                        import pyarrow.parquet as pq  # type:ignore
                    except ImportError as err:  # pragma: no cover
                        raise MissingDependencyError(
                            "`pyarrow` is missing, please install or includein requirements.txt"
                        )

                    import io
                    from functools import reduce

                    tempfile = io.BytesIO()

                    # convert to a list of dicts
                    temp_list = [
                        orjson.loads(record) for record in self.buffer.splitlines()
                    ]

                    # When writing to Parquet, the table gets the schema from the first
                    # row, if this row is missing columns (shouldn't, but it happens)
                    # it will be missing for all records, so get the columns from the
                    # entire dataset and ensure all records have the same columns.

                    # first, we get all the columns, from all the records
                    columns = reduce(
                        lambda x, y: x + [a for a in y.keys() if a not in x],
                        temp_list,
                        [],
                    )

                    # then we make sure each row has all the columns
                    temp_list = [
                        {column: row.get(column) for column in columns}
                        for row in temp_list
                    ]

                    pytable = pyarrow.Table.from_pylist(temp_list)
                    pyarrow.parquet.write_table(
                        pytable, where=tempfile, compression="zstd"
                    )

                    tempfile.seek(0)
                    self.buffer = tempfile.read()

                if self.format == "zstd":
                    # zstandard is an non-optional installed dependency
                    self.buffer = zstandard.compress(self.buffer)

                committed_blob_name = self.inner_writer.commit(
                    byte_data=bytes(self.buffer), override_blob_name=None
                )

                if "BACKOUT" in committed_blob_name:
                    get_logger().warning(
                        f"{self.records_in_buffer:n} failed records written to BACKOUT partition `{committed_blob_name}`"
                    )
                get_logger().debug(
                    {
                        "committed_blob": committed_blob_name,
                        "records": self.records_in_buffer,
                        "bytes": len(self.buffer),
                    }
                )
            finally:
                lock.release()

        self.buffer = bytearray()
        return committed_blob_name

    def open_buffer(self):
        self.buffer = bytearray()
        self.records_in_buffer = 0

    def __del__(self):
        # this should never be relied on to save data
        self.commit()



# cython: language_level=3
"""
This is an contained reading pipeline, intended to be run in a thread/processes.

The processing pipeline is made of a set of steps operating on a defined and finite
dataset:

┌────────────┬────────────────────────────────────────────────────────────┐
│ Function   │ Role                                                       │
├────────────┼────────────────────────────────────────────────────────────┤
│ Pre-Filter │ Pre-filter the rows based on a subset of the filters       │
│            │                                                            │
│ Read       │ Read the raw content from the file                         │
│            │                                                            │
│ Decompress │ Convert the raw content to lined data                      │
│            │                                                            │
│ Parse      │ Interpret the lined data into dictionaries                 │
│            │                                                            │
│ Filter     │ Apply full set of row filters to the read data             │
│            │                                                            │
│ Reduce     │ Aggregate                                                  │
└────────────┴────────────────────────────────────────────────────────────┘
"""
from mabel import logging
from . import decompressors, parsers
from enum import Enum
from functools import reduce
from ....utils import paths
from ....data.internals.records import flatten
from ....data.internals.index import Index
from ....data.internals.expression import Expression
from ....data.internals.dnf_filters import DnfFilters

logger = logging.get_logger()


def empty_list(x):
    return []


class EXTENSION_TYPE(str, Enum):
    # labels for the file extentions
    DATA = "DATA"
    CONTROL = "CONTROL"
    INDEX = "INDEX"


KNOWN_EXTENSIONS = {
    ".txt": (decompressors.block, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".json": (decompressors.block, parsers.json, EXTENSION_TYPE.DATA),
    ".zstd": (decompressors.zstd, parsers.json, EXTENSION_TYPE.DATA),
    ".lzma": (decompressors.lzma, parsers.json, EXTENSION_TYPE.DATA),
    ".zip": (decompressors.unzip, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".jsonl": (decompressors.lines, parsers.json, EXTENSION_TYPE.DATA),
    ".xml": (decompressors.block, parsers.xml, EXTENSION_TYPE.DATA),
    ".lxml": (decompressors.lines, parsers.xml, EXTENSION_TYPE.DATA),
    ".parquet": (decompressors.parquet, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".csv": (decompressors.csv, parsers.pass_thru, EXTENSION_TYPE.DATA),
    ".ignore": (empty_list, empty_list, EXTENSION_TYPE.CONTROL),
    ".complete": (empty_list, empty_list, EXTENSION_TYPE.CONTROL),
    ".idx": (empty_list, empty_list, EXTENSION_TYPE.INDEX),
}


pass_thru = lambda x: x


def no_filter(x):
    return True


def expand_nested_json(row):
    # this is really slow - on a simple read it's roughly 60% of the execution
    if hasattr(row, "items"):
        for k, v in [(k, v) for k, v in row.items()]:
            if hasattr(v, "items"):
                if hasattr(row, "as_dict"):
                    row = row.as_dict()  # only convert if we really have to
                row.update(flatten(dictionary={k: v}, separator="."))
                row.pop(k)
    return row


class ParallelReader:

    NOT_INDEXED = {-1}

    def __init__(
        self,
        reader,
        filters=no_filter,
        columns="*",
        reducer=pass_thru,
        override_format=None,
        **kwargs,
    ):
        """

        Parameters:
            reader: callable
            columns: callable
            filters: callable
            reducer: callable
            **kwargs: kwargs
        """

        # DNF form is a representation of logical expressions which it is easier
        # to apply any indicies to - we can convert Expressions to DNF but we can't
        # convert functions to DNF.
        if isinstance(filters, DnfFilters):
            self.dnf_filter = filters
        elif isinstance(filters, Expression):
            self.dnf_filter = DnfFilters(filters.to_dnf())
        else:
            self.dnf_filter = None

        # the reader gets the data from the storage platform e.g. read the file from
        # disk or download the file
        self.reader = reader

        self.columns = columns

        # this is the filter of the collected data, this can be used
        # against more operators
        self.filters = filters

        # this is aggregation and reducers for the data
        self.reducer = reducer

        # sometimes the user knows better
        self.override_format = override_format

        if self.override_format:
            self.override_format = self.override_format.lower()
            if not self.override_format[0] == ".":
                self.override_format = "." + self.override_format

    def pre_filter(self, blob_name, index_files):
        """
        Select rows from the file based on the filters and indexes, this filters
        the data before we've even seen it. This is usually faster as it avoids
        reading files with no records and parsing data we don't want.

        Go over the tree, if the predicate is an operator we can filter on and
        the field in the predicate is indexed, we can apply a pre-filter.

        ORs are fatal if both sides can't be pre-evaluated.
        """

        SARGABLE_OPERATORS = {"=", "==", "is", "in", "contains"}

        def _inner_prefilter(predicate):
            # No filter doesn't filter
            if predicate is None:  # pragma: no cover
                return None

            # If we have a tuple, this is the inner most representation of our filter.
            # Extract out the key, operator and value and look them up against the
            # index - if we can.
            if isinstance(predicate, tuple):
                key, operator, values = predicate
                if operator in SARGABLE_OPERATORS:
                    # do I have an index for this field?
                    for index_file in [
                        index_file
                        for index_file in index_files
                        if f".{key}." in index_file
                    ]:
                        index = Index(self.reader.get_blob_stream(index_file))
                        return index.search(values)

                return self.NOT_INDEXED

            if isinstance(predicate, list):
                # Are all of the entries tuples? These are ANDed together.
                rows = []
                if all([isinstance(p, tuple) for p in predicate]):
                    for index, row in enumerate(predicate):
                        if index == 0:
                            rows = _inner_prefilter(row)
                        else:
                            rows = [p for p in _inner_prefilter(row) if p in rows]
                    return rows

                # Are all of the entries lists? These are ORed together.
                # All of the elements in an OR need to be indexable for use to be
                # able to prefilter.
                if all([isinstance(p, list) for p in predicate]):
                    evaluations = [_inner_prefilter(p) for p in predicate]
                    if not all([e == self.NOT_INDEXED for e in evaluations]):
                        return self.NOT_INDEXED
                    else:
                        # join the data sets together by adding them
                        rows = reduce(
                            lambda x, y: x + _inner_prefilter(y), predicate, []
                        )
                    return rows

                # if we're here the structure of the filter is wrong
                return self.NOT_INDEXED
            return self.NOT_INDEXED

        index_filters = _inner_prefilter(self.dnf_filter.predicates)
        if index_filters != self.NOT_INDEXED:
            return True, index_filters
        return False, []

    def _select(self, data_set, selector):
        for index, record in enumerate(data_set):
            if index in selector:
                yield record

    def __call__(self, blob_name, index_files):
        # print(blob_name, "in")
        try:
            if self.override_format:
                ext = self.override_format
            else:
                bucket, path, stem, ext = paths.get_parts(blob_name)

            if ext not in KNOWN_EXTENSIONS:
                return []
            decompressor, parser, file_type = KNOWN_EXTENSIONS[ext]

            # Pre-Filter
            if self.dnf_filter and len(index_files):
                (row_selector, selected_rows) = self.pre_filter(blob_name, index_files)
            else:
                row_selector = False
                selected_rows = []

            # if the value isn't in the set, abort early
            if row_selector and len(selected_rows) == 0:
                return None

            # Read
            record_iterator = self.reader.read_blob(blob_name)
            # Decompress
            record_iterator = decompressor(record_iterator)
            ### bypass rows which aren't selected
            if row_selector:
                record_iterator = self._select(record_iterator, selected_rows)
            # Parse
            record_iterator = map(parser, record_iterator)
            # Expand Nested JSON
            # record_iterator = map(expand_nested_json, record_iterator)
            # Transform
            record_iterator = map(self.columns, record_iterator)
            # Filter
            record_iterator = filter(self.filters, record_iterator)
            # Reduce
            record_iterator = self.reducer(record_iterator)
            # Yield
            yield from record_iterator
            # print(blob_name, "out")
        except Exception as e:
            import traceback

            logger.error(f"{blob_name} had an error - {e}\n{traceback.format_exc()}")
            return []



"""
StorageClassCompressedMemory is a helper class for persisting DictSets locally, it is
the backend for the COMPRESSED_MEMORY variation of the STORAGE CLASSES.

This is particularly useful when local storage is slow and the data is too big for
memory, or you are doing a lot of operations on the data.

If you're doing few operations on the the data and it is easily recreated or local
storage is fast, this isn't a good option.
"""
import orjson
import lz4.frame
from itertools import zip_longest
from . import BaseStorageClass

BATCH_SIZE = 500


class StorageClassCompressedMemory(BaseStorageClass):
    def __init__(self, iterable):
        compressor = lz4.frame

        self.batches = []
        self.length = 0

        batch = None
        for batch in zip_longest(*[iterable] * BATCH_SIZE):
            self.length += len(batch)
            # pay for fast deserialization here:
            if hasattr(batch[0], "mini"):
                json_batch = (
                    b"["
                    + b",".join([item.mini for item in batch if not item is None])
                    + b"]"
                )
                self.batches.append(compressor.compress(json_batch))
            else:
                self.batches.append(compressor.compress(orjson.dumps(batch)))

        # the last batch fills with Nones
        if batch:
            self.length -= batch.count(None)
        self.iterator = None

    def _inner_reader(self, *locations):
        decompressor = lz4.frame

        if locations:
            if not isinstance(locations, (tuple, list, set)):
                locations = [locations]
            ordered_location = sorted(locations)
            batch_number = -1
            batch = []
            for i in ordered_location:
                requested_batch = i % BATCH_SIZE
                if requested_batch != batch_number:
                    batch = self.parse_json(
                        decompressor.decompress(self.batches[requested_batch])
                    )
                yield batch[i - i % BATCH_SIZE]

        else:
            for batch in self.batches:
                records = self.parse_json(decompressor.decompress(batch))
                for record in records:
                    if record:
                        yield record

    def __iter__(self):
        self.iterator = iter(self._inner_reader())
        return self.iterator

    def __next__(self):
        if not self.iterator:
            self.iterator = iter(self._inner_reader())
        return next(self.iterator)

    def __len__(self):
        return self.length



"""
Google Cloud Storage Reader
"""
import io
import os
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...utils import paths


class GoogleStorageReadError(Exception):
    pass


def url_encoder(url):
    result_url = ""
    for character in url:
        if not character.isalnum() and character not in ("-", "."):
            result_url += "%" + hex(ord(character))[2:]
        else:
            result_url += character
    return result_url


class GoogleCloudStorageReader(BaseInnerReader):

    RULES = [
        {"name": "project", "required": False},
        {"name": "credentials", "required": False},
    ]

    def __init__(self, project: str, credentials=None, **kwargs):
        super().__init__(**kwargs)
        self.project = project
        self.credentials = credentials

    def get_blob_stream(self, blob_name):
        bucket, object_path, name, extension = paths.get_parts(blob_name)
        blob = get_blob(
            project=self.project,
            bucket=bucket,
            blob_name=object_path + name + extension,
            credentials=self.credentials,
        )
        return blob

    def get_blob_chunk(self, blob_name: str, start: int, buffer_size: int) -> bytes:
        bucket, object_path, name, extension = paths.get_parts(blob_name)
        blob = get_blob(
            project=self.project,
            bucket=bucket,
            blob_name=object_path + name + extension,
            credentials=self.credentials,
        )
        stream = blob.download_as_bytes(
            start=start, end=min(blob.size, start + buffer_size - 1)
        )
        return stream

    def get_blobs_at_path(self, path):
        bucket, object_path, name, extension = paths.get_parts(path)

        import requests

        # determin the domain
        domain = os.environ.get(
            "STORAGE_EMULATOR_HOST", "https://storage.googleapis.com"
        )
        if domain[-1] != "/":
            domain += "/"

        # add the headers if needed
        headers = {}
        if self.credentials:
            headers["Authorization"] = f"Bearer {self.credentials}"

        # get the data
        payload = requests.get(
            url=f"{domain}storage/v1/b/{url_encoder(bucket)}/o?prefix={object_path}",
            headers=headers,
            timeout=30,
        )

        if payload.status_code // 100 != 2:
            return []

        yield from [
            bucket + "/" + blob["name"]
            for blob in payload.json()["items"]
            if not blob["name"].endswith("/")
        ]


def get_blob(project: str, bucket: str, blob_name: str, credentials=None):

    import requests

    # determin the domain
    domain = os.environ.get("STORAGE_EMULATOR_HOST", "https://storage.googleapis.com")
    if domain[-1] != "/":
        domain += "/"

    # add the headers if needed
    headers = {}
    if credentials:
        headers["Authorization"] = f"Bearer {credentials}"

    # get the data
    payload = requests.get(
        url=f"{domain}storage/v1/b/{bucket}/o/{blob_name}?alt=media",
        headers=headers,
        timeout=30,
    )

    if payload.status_code // 100 != 2:
        raise GoogleStorageReadError(payload.content)

    return io.BytesIO(payload.content)



from abc import ABC, abstractclassmethod

try:
    import simdjson

    def json(ds):
        """parse each line in the file to a dictionary"""
        json_parser = simdjson.Parser()
        return json_parser.parse(ds)


except ImportError:
    import orjson

    def json(ds):
        return orjson.loads(ds)


class BaseStorageClass(ABC):

    iterator = None
    length = -1

    def __init__(self, iterator):
        raise NotImplementedError()

    def _inner_reader(self, *locations):
        raise NotImplementedError()

    def __iter__(self):
        self.iterator = self._inner_reader()
        return self.iterator

    def __next__(self):
        if not self.iterator:
            self.iterator = self._inner_reader()
        return next(self.iterator)

    def __len__(self):
        return self.length

    def parse_json(self, ds):
        return json(ds)



"""
Base Inner Reader
"""
from functools import lru_cache
import io
import abc
import pathlib
import datetime
from io import IOBase
from typing import Iterable
from mabel.utils import paths, dates
from mabel.logging import get_logger


BUFFER_SIZE: int = 64 * 1024 * 1024  # 64Mb


@lru_cache(1)
def memcached_server():
    import os

    # the server must be set in the environment
    memcached_config = os.environ.get("MEMCACHED_SERVER", None)
    if memcached_config is None:
        return None

    # expect either SERVER or SERVER:PORT entries
    memcached_config = memcached_config.split(":")
    if len(memcached_config) == 1:
        # the default memcached port
        memcached_config.append(11211)

    # we need the server and the port
    if len(memcached_config) != 2:
        return None

    try:
        from pymemcache.client import base
    except ImportError:
        return None

    # wait 1 second to try to connect, it's not worthwhile as a cache if it's slow
    return base.Client(
        (
            memcached_config[0],
            memcached_config[1],
        ),
        connect_timeout=1,
        timeout=1,
    )


class BaseInnerReader(abc.ABC):
    def _extract_as_at(self, path):
        parts = path.split("/")
        for part in parts:
            if part.startswith("as_at_"):
                return part
        return ""

    def _extract_by(self, path):
        parts = path.split("/")
        for part in parts:
            if part.startswith("by_"):
                return part

    def __init__(self, partitions=None, partition_filter=None, **kwargs):
        self.dataset = kwargs.get("dataset")
        if self.dataset is None:
            raise ValueError("Readers must have the `dataset` parameter set")
        if not self.dataset.endswith("/"):
            self.dataset += "/"
        # if "{" not in self.dataset and not kwargs.get("raw_path", False):
        #    self.dataset += "{datefolders}/"
        if partitions:
            self.dataset += "/".join(partitions) + "/"
        self.partition_filter = partition_filter

        start_date = dates.extract_date(kwargs.get("start_date"))
        end_date = dates.extract_date(kwargs.get("end_date"))

        self.start_date = min(start_date, end_date)
        self.end_date = max(start_date, end_date)

        self.days_stepped_back = 0

    def step_back_a_day(self):
        """
        Steps back a day so data can be read from a previous day
        """
        self.days_stepped_back += 1
        self.start_date -= datetime.timedelta(days=1)
        self.end_date -= datetime.timedelta(days=1)
        return self.days_stepped_back

    def __del__(self):
        """
        Only here in case a helpful dev expects te base class to have it
        """
        pass

    @abc.abstractmethod
    def get_blobs_at_path(self, prefix=None) -> Iterable:
        pass

    @abc.abstractmethod
    def get_blob_bytes(self, blob: str) -> bytes:
        """
        Return a filelike object
        """
        pass

    def read_blob(self, blob: str) -> IOBase:
        """
        Read-thru cache
        """
        cache_server = memcached_server()
        # if cache isn't configured, read and get out of here
        if not cache_server:
            result = self.get_blob_bytes(blob)
            return io.BytesIO(result)

        # hash the blob name for the look up
        from siphashc import siphash

        blob_hash = str(siphash("RevengeOfTheBlob", blob))

        # try to fetch the cached file
        result = cache_server.get(blob_hash)

        # if the item was a miss, get it from storage and add it to the cache
        if result is None:
            result = self.get_blob_bytes(blob)
            cache_server.set(blob_hash, result)

        return io.BytesIO(result)

    def get_list_of_blobs(self):

        blobs = []
        # For each day in the range, get the blobs for us to read
        for cycle_date in dates.date_range(self.start_date, self.end_date):
            # Build the path name
            cycle_path = pathlib.Path(
                paths.build_path(path=self.dataset, date=cycle_date)
            )
            cycle_blobs = list(self.get_blobs_at_path(path=cycle_path))

            # Remove any BACKOUT data - this is essentially a DEAD LETTER queue
            # so we don't want to include in when reading
            cycle_blobs = [blob for blob in cycle_blobs if "BACKOUT" not in blob]

            # The partitions are stored in folders with the prefix 'by_', as in,
            # partitioned **by** field name
            list_of_partitions = {
                self._extract_by(blob) for blob in cycle_blobs if "/by_" in blob
            }

            # If we've been provided a partition_filter search hint, try to use this
            # first to prune data
            chosen_partition = ""

            if self.partition_filter:
                from mabel.utils import text

                # break the filter into parts, and make sure they're safe and valid
                (
                    partition_filter_field,
                    partition_filter_op,
                    partition_filter_value,
                ) = self.partition_filter
                if partition_filter_op not in ("=", "=="):
                    raise NotImplementedError(
                        "`partition_filter` operation can only be equals (`=`)"
                    )
                partition_filter_field = text.sanitize(partition_filter_field)
                partition_filter_value = text.sanitize(partition_filter_value)
                partition_filter = f"/by_{partition_filter_field}/{partition_filter_field}={partition_filter_value}/"

                # If we can find the partition in the folder set, then prune to it
                if any(
                    [f"by_{partition_filter_field}" in by for by in list_of_partitions]
                ):
                    # Do the pruning
                    cycle_blobs = [
                        blob for blob in cycle_blobs if partition_filter in blob
                    ]
                    #  We only have one partition now
                    list_of_partitions = [f"by_{partition_filter_field}"]
                    get_logger().debug(
                        f"Applied partition filter by: `{partition_filter}`"
                    )
                else:
                    get_logger().debug(
                        f"Wasn't able to find partition to filter by: `{partition_filter}`"
                    )

            # If we have multiple 'by_' partitions, pick one (pick the first one)
            if list_of_partitions:
                list_of_partitions = sorted(list_of_partitions)
                chosen_partition = list_of_partitions.pop()
                if list_of_partitions:
                    get_logger().info(
                        f"Ignoring {len(list_of_partitions)} 'by' partitionings, reading from '{chosen_partition}'"
                    )
                # Do the pruning
                cycle_blobs = [
                    blob for blob in cycle_blobs if f"/{chosen_partition}/" in blob
                ]

            def safe_get_next(lst, item):
                try:
                    index = lst.index(item)
                    return lst[index + 1]
                except:
                    return None

            # Cycle over the list of partitions (e.g. the hour=02 bits) we can't use
            # the frame id of one on the rest
            if chosen_partition == "":
                partitioned_folders = {""}
            else:
                partitioned_folders = {
                    safe_get_next(blob.split("/"), chosen_partition)
                    for blob in cycle_blobs
                }

            for partitioned_folder in partitioned_folders:

                partitioned_blobs = [
                    blob
                    for blob in cycle_blobs
                    if f"{chosen_partition}/{partitioned_folder}" in blob
                ]

                # Work out if there's an as_at part
                as_ats = {
                    self._extract_as_at(blob)
                    for blob in partitioned_blobs
                    if "as_at_" in blob
                }
                if as_ats:
                    as_ats = sorted(as_ats)
                    as_at = as_ats.pop()

                    is_complete = lambda blobs: any(
                        [blob for blob in blobs if as_at + "/frame.complete" in blob]
                    )
                    is_invalid = lambda blobs: any(
                        [blob for blob in blobs if (as_at + "/frame.ignore" in blob)]
                    )

                    while not is_complete(partitioned_blobs) or is_invalid(
                        partitioned_blobs
                    ):
                        if not is_complete(partitioned_blobs):
                            get_logger().debug(
                                f"Frame `{partitioned_folder}/{as_at}` is not complete - `frame.complete` file is not present - skipping this frame."
                            )
                        if is_invalid(partitioned_blobs):
                            get_logger().debug(
                                f"Frame `{partitioned_folder}/{as_at}` is invalid - `frame.ignore` file is present - skipping this frame."
                            )
                        if len(as_ats) > 0:
                            as_at = as_ats.pop()
                        else:
                            return []
                    get_logger().debug(f"Reading from DataSet frame `{as_at}`")
                    partitioned_blobs = [
                        blob
                        for blob in partitioned_blobs
                        if (as_at in blob) and ("/frame.complete" not in blob)
                    ]

                blobs += partitioned_blobs

        return sorted(blobs)



"""
Index is a json KV store:

{ 
    "hash(value)": [list of locations]
}
"""
from typing import Iterable
import orjson
from siphashc import siphash

MAX_INDEX = 4294967295  # 2^32 - 1
SEED = "eschatologically"  # needs to be 16 characters long

"""
There are overlapping terms because we're traversing a dataset so we can traverse a
dataset. 

Terminology:
    Entry     : a record in the Index
    Position  : the position of the entry in the Index
    Location  : the position of the row in the target file
    Row       : a record in the target file
"""


class Index:
    def __init__(self, index: bytes):
        if hasattr(index, "read"):
            index = index.read()  # type:ignore

        try:
            import simdjson

            self._index = simdjson.Parser().parse(index)
        except ImportError:
            self._index = orjson.loads(index)

    @staticmethod
    def build_index(dictset: Iterable[dict], column_name: str):
        """
        Build an index from a dictset.

        Parameters:
            dictset: iterable of dictionaries
                The dictset to index
            column_name: string
                The name of the index which will be indexed

        Returns:
            bytes
        """
        # We do this in two-steps
        # 1) Build an intermediate form of the index as a list of entries
        # 2) Conver that intermediate form into a binary index
        builder = IndexBuilder(column_name)
        for position, row in enumerate(dictset):
            builder.add(position, row)
        return builder.build()

    def search(self, search_term) -> Iterable:
        """
        Search the index for a value. Returns a list of row numbers, if the value is
        not found, the list is empty.
        """
        if not isinstance(search_term, (list, set, tuple)):
            search_term = [search_term]
        result: list = []
        for term in search_term:
            key = format(siphash(SEED, f"{term}") % MAX_INDEX, "x")
            if key in self._index:  # type:ignore
                result[0:0] = self._index[key]  # type:ignore
        return result

    def dump(self, file):
        with open(file, "wb") as f:
            f.write(self.bytes())

    def bytes(self):
        if hasattr(self._index, "mini"):
            return self._index.mini
        import orjson

        return orjson.dumps(self._index)


class IndexBuilder:

    slots = ("column_name", "temporary_index")

    def __init__(self, column_name: str):
        self.column_name: str = column_name
        self.temporary_index: Iterable[dict] = []

    def add(self, position, record):
        ret_val = []
        if record.get(self.column_name):
            # index lists of items separately
            values = record[self.column_name]
            if not isinstance(values, list):
                values = [values]
            for value in values:
                entry = (format(siphash(SEED, f"{value}") % MAX_INDEX, "x"), position)
                ret_val.append(entry)
        self.temporary_index += ret_val
        return ret_val

    def build(self) -> Index:
        temp_index = {}
        for val, pos in self.temporary_index:
            if not val in temp_index:
                temp_index[val] = [pos]
            else:
                poses = temp_index[val]
                poses.append(pos)
                temp_index[val] = poses
        import orjson

        return Index(orjson.dumps(temp_index))



"""
This allows us to Monkey Patch types that aren't usually Monkey Patchable
"""
import ctypes
from types import MappingProxyType

# figure out side of _Py_ssize_t
if hasattr(ctypes.pythonapi, "Py_InitModule4_64"):
    _Py_ssize_t = ctypes.c_int64
else:
    _Py_ssize_t = ctypes.c_int  # type:ignore

# regular python
class _PyObject(ctypes.Structure):
    pass


_PyObject._fields_ = [
    ("ob_refcnt", _Py_ssize_t),
    ("ob_type", ctypes.POINTER(_PyObject)),
]

# python with trace
if object.__basicsize__ != ctypes.sizeof(_PyObject):

    class _PyObject(ctypes.Structure):  # type:ignore
        pass

    _PyObject._fields_ = [
        ("_ob_next", ctypes.POINTER(_PyObject)),
        ("_ob_prev", ctypes.POINTER(_PyObject)),
        ("ob_refcnt", _Py_ssize_t),
        ("ob_type", ctypes.POINTER(_PyObject)),
    ]


class _DictProxy(_PyObject):
    _fields_ = [("dict", ctypes.POINTER(_PyObject))]


def reveal_dict(proxy):
    if not isinstance(proxy, MappingProxyType):
        raise TypeError("dictproxy expected")
    dp = _DictProxy.from_address(id(proxy))
    ns = {}
    ctypes.pythonapi.PyDict_SetItem(
        ctypes.py_object(ns), ctypes.py_object(None), dp.dict
    )
    return ns[None]


def get_class_dict(cls):
    d = getattr(cls, "__dict__", None)
    if d is None:
        raise TypeError("given class does not have a dictionary")
    if isinstance(d, MappingProxyType):
        return reveal_dict(d)
    return d


def patch(cls, attr, value):
    cd = get_class_dict(cls)
    cd[attr] = value



from setuptools import setup, find_packages
from Cython.Build import cythonize

with open("mabel/version.py", "r") as v:
    vers = v.read()
exec(vers)  # nosec

with open("README.md", "r") as rm:
    long_description = rm.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()


ext_modules = cythonize(
    [
        "mabel/data/internals/group_by.py",
        "mabel/data/internals/dictset.py",
        "mabel/data/internals/expression.py",
        "mabel/data/readers/internals/inline_evaluator.py",
        "mabel/data/readers/internals/parallel_reader.py",
    ]
)

setup(
    name="mabel",
    version=__version__,
    description="Python Data Libraries",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Joocer",
    author="joocer",
    author_email="justin.joyce@joocer.com",
    packages=find_packages(include=["mabel", "mabel.*"]),
    url="https://github.com/mabel-dev/mabel/",
    install_requires=required,
    ext_modules=ext_modules,
)



"""
Separate the implementation of the Cursor from the Reader.

Cursor is made of three parts:
- map      : a bit array representing all of the blobs in the set - unread blobs
             are 0s and read blobs are 1s. This allows for blobs to be read in 
             an arbitrary order - although currently only implemented linearly.
- partition: the active parition (blob) that is being read
- location : the record in the active partition (blob), so we can resume reading
             midway through the blob if required.
"""
from siphashc import siphash
import orjson


class InvalidCursor(Exception):
    pass


class Cursor:
    def __init__(self, readable_blobs, cursor=None):
        # sort the readable blobs so they are in a consistent order
        self.readable_blobs = sorted(readable_blobs)
        self.read_blobs = []
        self.partition = ""
        self.location = -1

        if cursor:
            self.load_cursor(cursor)

    def load_cursor(self, cursor):
        from bitarray import bitarray

        if cursor is None:
            return

        if isinstance(cursor, str):
            cursor = orjson.loads(cursor)

        if (
            not "location" in cursor.keys()
            or not "map" in cursor.keys()
            or not "partition" in cursor.keys()
        ):
            raise InvalidCursor(f"Cursor is malformed or corrupted {cursor}")

        self.location = cursor["location"]
        find_partition = [
            blob
            for blob in self.readable_blobs
            if siphash("%" * 16, blob) == cursor["partition"]
        ]
        if len(find_partition) == 1:
            self.partition = find_partition[0]
        map_bytes = bytes.fromhex(cursor["map"])
        blob_map = bitarray()
        blob_map.frombytes(map_bytes)
        self.read_blobs = [
            self.readable_blobs[i]
            for i in range(len(self.readable_blobs))
            if blob_map[i]
        ]

    def next_blob(self, previous_blob=None):
        if previous_blob:
            self.read_blobs.append(previous_blob)
            self.partition = ""
            self.location = -1
        if self.partition and self.location >= 0:
            if self.partition in self.readable_blobs:
                return self.partition
            partition_finder = [
                blob
                for blob in self.readable_blobs
                if siphash("%" * 16, blob) == self.partition
            ]
            if len(partition_finder) != 1:
                raise ValueError(
                    f"Unable to determine current partition ({self.partition})"
                )
            return partition_finder[0]
        unread = [blob for blob in self.readable_blobs if blob not in self.read_blobs]
        if len(unread) > 0:
            self.partition = unread[0]
            self.location = -1
            return self.partition
        return None

    def skip_to_cursor(self, iterator):
        # cycle through the iterator to the cursor location
        if self.location < 0:
            return 0
        for index in range(self.location + 1):
            next(iterator, None)
        return self.location

    def get(self):
        return {
            "map": self["map"],
            "partition": self["partition"],
            "location": self["location"],
        }

    def __getitem__(self, item):
        from bitarray import bitarray

        if item == "map":
            blob_map = bitarray(
                "".join(
                    [
                        "1" if blob in self.read_blobs else "0"
                        for blob in self.readable_blobs
                    ]
                )
            )
            return blob_map.tobytes().hex()
        if item == "partition":
            return siphash("%" * 16, self.partition)
        if item == "location":
            return self.location
        return None

    def __repr__(self):
        return orjson.dumps(self.get()).decode("UTF8")
